package Source::Repository;

use strict;
use warnings;
use Carp;

use File::Path qw(make_path remove_tree);
use File::Basename;
use Source::Package;

our $VERSION=0.001;
our $verbose=0;

sub new {
    my $proto=shift;
    my $class=ref($proto) || $proto;
    my $self = {
	-cachedir=>'in',
	-path=>[],
	-names=>[],
	@_,
    };
    my $default_opt=$self->{-default_opt};
    my $default_opt_is_invalid;
    if ($default_opt) {
	print STDERR "Source::Repository: default_opt=$default_opt\n" if $verbose > 3;
	my @path=grep {-d $_} &Source::Repository::path_to_array($default_opt);
	unless (-f $default_opt or @path) {
	    $default_opt_is_invalid=1;
	    warn "Source::Repository: not a valid path(s) or file: $default_opt\n" if $verbose;
	}
	if (not @{$self->{-path}} and not $self->{-pathlist} and @path) {
	    $self->{-path}=\@path;
	} elsif (not $self->{-repolist} and -f $default_opt) {
	    $self->{-repolist}=$default_opt;
	} else {
	    warn "Source::Repository: already configured - ignored default_opt $default_opt\n" if $verbose;
	}
    }
    if ($self->{-pathlist}) {
	push @{$self->{-path}}, &Source::Repository::path_to_array($self->{-pathlist});
    }
    bless($self,$class);
    $self->__load_repolist();
    $self->__load_namelist();
    confess "internal error: invalid repository path" if not ref $self->{-path};
    confess "internal error: invalid list of names" if not ref $self->{-names};

    if (not $self->{-name2pkg} and not scalar @{$self->{-path}}) {
	if ($default_opt_is_invalid) {
	    die "Source::Repository: can't initialize repository: not a valid path(s) or file: $default_opt\n";
	} else {
	    confess "Source::Repository: can't initialize repository: neither path nor list is specified\n";
	}
    }
    return $self;
}


sub expand_glob {
    my ($self,$item)=@_;
    return &__expand_glob($item,$self->{-names});
}

# used in lib/Source/Repository/Mass.pm for --all option
# TODO: implement in RPM
sub names {
    my ($self)=@_;
    return @{$self->{-names}};
}

# can be redefined in subclasses to allow download by name
sub download {
    my ($self,$pkg)=@_;
    ...;
}

sub is_pkg {
   my ($self,$item)=@_;
   return defined $item and -e $item and not -d $item;
}

sub name2pkg {
    my ($self,$name)=@_;
    my $name2pkg=$self->{-name2pkg};
    if ($name2pkg) {
	return $name2pkg->{$name};
    } elsif ($self->{-findsub}) {
	my $file=&{$self->{-findsub}}($name,@{$self->{-path}});
	return unless $file;
	return $self->file2pkg();
    }
    return;
}

sub __load_repolist {
    my $self=shift;
    my $list=$self->{-repolist};
    return unless defined $list;
    ($self->{-names}, $self->{-name2pkg})=&__load_list($list,\&default_list3_processor,$self);

}

sub __load_namelist {
    my $self=shift;
    my $list=$self->{-namelist};
    return unless defined $list;
    ($self->{-names})=&__load_list($list);
}

#------------------------------------------
#    protected utilities
#------------------------------------------

# used in Source::Repository::Mass and Source::Repository::RPM::ALTLinuxSrcList
sub __line2array {
    my $line=shift;
    chomp $line;
    $line=~s/^\s*//;
    $line=~s/\s*$//;
    return if $line=~/^\s*#/;
    return if $line eq '';
    return split (/\s+/, $line);
}

# used in RPM
sub path_to_array {
    my $paths=shift;
    my @path;
    foreach (split(/(?<!\\):/,$paths)) {
	# support for paths like widehat.opensuse.org/repositories/games\:/unknown-horizons
	s!\\:!:!g;
	if ($_ eq '') {
	    $_ = '.';
	} elsif (/^~/) {
	    s/^~/$ENV{HOME}/;
	}
	#warn "path_to_array: [$_]\n";
	push @path, $_ if -d $_;
    }
    return @path;
}

#------------------------------------------
#    private utilities
#------------------------------------------

# glob expansion on $opt_namelist
sub __expand_glob {
    my ($pattern, $arrayptr)=@_;
    return $pattern unless $pattern =~/[?*]/;
    my $regexp=&glob2pat($pattern);
    my @names = grep {/$regexp/} @$arrayptr;
    print STDERR "WARNING: pattern $pattern does not match.\n" unless @names;
    return @names;
}

sub glob2pat {
    my $globstr = shift;
    my %patmap = (
	'*' => '.*',
	'?' => '.',
	'[' => '[',
	']' => ']',
	'{' => '(?:',
	'}' => ')',
	',' => '|',
	);
    $globstr =~ s{(.)} { $patmap{$1} || "\Q$1" }ge;
    return '^' . $globstr . '$';
}

sub __load_list {
    my $list=shift;
    my $processor=shift;
    return unless defined $list;
    my @all_names;
    my %name2pkg;
    open (my $fn, '<', $list) or die "can't open file $list\n";
    while (my $line=<$fn>) {
	my @list = &__line2array($line);
	next unless @list;
	my $name=$list[0];
	push @all_names, $name;
	$name2pkg{$name}=&$processor(\@list,@_) if $processor;
    }
    close($fn);
    return \@all_names, \%name2pkg;
}

sub default_list3_processor {
    my ($ref,$self)=@_;
    return unless $ref;
    my ($epoch,$release);
    return unless @$ref;
    my ($name, $versionspec, $url)=@$ref;
    $epoch=$1 if $versionspec=~s/^(\d+)://;
    # warning: altlinux-specific, for src.list only
    $release=$1 if $versionspec=~s/-(alt.+)$//;
    # TODO: baseurl support
    return Source::Package->new(
	-name => $name,
	-epoch=> $epoch,
	-version => $versionspec,
	-release => $release,
	-url => $url,
	);
}
1;
