package Source::Repository::Mass;

use strict;
use warnings;
use Cwd;
use Carp;
use Getopt::Long;
use File::Basename;
use File::Path qw(make_path remove_tree);
use RPM qw/evrcmp vercmp/;

use RPM::Source::Shared::ResourceContainer;
use RPM::Source::Shared::Resource::Debug;
use Source::Package::Pair;
use Source::Package::Comparator;
use Source::Package::Comparators;
use Source::Repository::Mass::PkgFilter;
use Source::Repository::RPM;
use Source::Repository::Matcher;
use Source::Repository::Filter;

require Exporter;

our $VERSION=0.12;

our @ISA = qw(Exporter);
our @EXPORT = qw/$DESTPATH $ORIGINPATH/;
our ($DESTPATH, $ORIGINPATH);

our $comparator_class='Source::Package::Comparator';
our $matcher_class='Source::Repository::Matcher';
our $origin_class='Source::Repository::RPM';
our $dest_class='Source::Repository::RPM';

our $epoch_comparator=\&Source::Package::Comparators::null_epoch_comparator;
our $version_comparator=\&Source::Package::Comparators::pure_rpm_version_comparator;
our $release_comparator=\&Source::Package::Comparators::pure_rpm_release_comparator;

our @ORIGIN_DEFAULT_OPTS;
our @DEST_DEFAULT_OPTS;

our $verbose=0;
our $strategy = 'all';

my ($opt_transaction, $opt_taskfile, $opt_all, $opt_destmode, $help);
my ($opt_origin_namelist, $opt_dest_namelist, $opt_namelist);
my ($opt_origin_repolist, $opt_dest_repolist);
my ($opt_origin_pathlist, $opt_dest_pathlist);
my ($opt_origin_cachedir, $opt_dest_cachedir);

our ($opt_skip_on_error);

our %STRATEGYMAP=(
    # quick synonyms
    'force' => \&origin2dest,
    'f' => \&origin2dest,
    'a' => \&origin2dest,
    'm' => \&originskipdest,
    'o' => \&destskiporigin,
    'u' => \&destup,
    'n' => \&newer,

    'all' => \&origin2dest,
    'missing' => \&originskipdest,
    'orphaned' => \&destskiporigin,
    'destup' => \&destup,
    'newer' => \&newer,
    'newolder' => \&newolder,
    'older' => \&older,
    'both' => \&both,
    # S(ilent) names
    'allS' => \&origin2destS,
    'missingS' => \&originskipdestS,
    'orphanedS' => \&destskiporiginS,
    'destupS' => \&destupS,
    'newerS' => \&newerS,
    'olderS' => \&olderS,
    'bothS' => \&bothS,

    # old names
    'origin2dest' => \&origin2dest,
    'originskipdest' => \&originskipdest,
    'destskiporigin' => \&destskiporigin,

    # verup, relup
    'verup' => \&verup,
    'relup' => \&relup,

    # print help
    'help' => \&help_strategy,
    );

our %STRATEGYDESCR=(
    'all' => 'any package',
    'missing' => 'packages in origin, missing in dest', 
    'orphaned' => 'packages in dest, missing in origin',
    'destup' => 'packages in dest, older than origin (i.e. need update)',
    'newer' => 'packages in origin, new and newer than dest',
    'newolder' => 'packages in origin, new and older than dest',
    'older' => 'packages in origin, older than dest',
    'both' => 'packages in origin, present in dest',
    'relup' => 'destup, limited release update',
    'verup' => 'destup, limited to version update',

    );

my @LONGOPT = (
	"all"	=> \$opt_all,
	"dest=s"	=> \$DESTPATH,
	"dest-cachedir=s"	=> \$opt_dest_cachedir,
	"dest-namelist=s"	=> \$opt_dest_namelist,
	"dest-repolist=s"	=> \$opt_dest_repolist,
	"dest-path=s"		=> \$opt_dest_pathlist,
	"input-dest"	=> sub {$opt_destmode=1},
	"input-origin"	=> sub {$opt_destmode=0},
	"origin=s"	=> \$ORIGINPATH,
	"origin-cachedir=s"	=> \$opt_origin_cachedir,
	"origin-namelist=s"	=> \$opt_origin_namelist,
	"origin-repolist=s"	=> \$opt_origin_repolist,
	"origin-path=s"		=> \$opt_origin_pathlist,
	"skip-on-error"	=> \$opt_skip_on_error,
	"strategy=s"	=> \$strategy,
	"transaction"	=> \$opt_transaction,
	"tasklist=s"	=> \$opt_taskfile,
	"namelist=s"	=> \$opt_namelist,

	"help"		=> \$help,
	"v|verbose+"	=> \$verbose,
	'q|quiet'   => sub { $verbose = 0 },
    );

our @EXT_OPTS;
our @EXT_OPTS_USAGE;
our @EXT_LONGOPT_CLASSES;

my @built_in_containers=(
    'RPM::Source::Shared::ResourceContainer', 
    'Source::Repository::Matcher', 
    'Source::Repository::Filter',
    'Source::Repository::Mass::PkgFilter');

sub new {
    my $class = shift;
    my $self = {
	# replace given name with $pkg->name query (can be time consuming)
	-use_pkg_name => 0,
	'-originopts' => [],
	'-destopts' => [],
	'-cliargs' => [],
	'-packages' => [],
	@_
    };
    bless $self, $class;
    my $result = GetOptions (
	@LONGOPT,
	@EXT_OPTS,
	@{$self->{'-cliargs'}},
	# should be always the last (under the map)
	map {$_->get_longopt} @built_in_containers, @EXT_LONGOPT_CLASSES);

    if (not $result) {
	print "run $0 --help for details.\n";
	exit 1;
    } elsif ($help) {
	$self->usage();
	exit 0;
    }

    map {$_->process_options} @built_in_containers, @EXT_LONGOPT_CLASSES;

    $self->{VERBOSE}=$verbose;

    my (@defrepo_opts, @origin_opts, @dest_opts);
    push @defrepo_opts, -namelist => $opt_namelist if $opt_namelist;
    push @origin_opts, -namelist => $opt_origin_namelist if $opt_origin_namelist;
    push @dest_opts, -namelist => $opt_dest_namelist if $opt_dest_namelist;
    push @origin_opts, -repolist => $opt_origin_repolist if $opt_origin_repolist;
    push @dest_opts, -repolist => $opt_dest_repolist if $opt_dest_repolist;
    push @origin_opts, -cachedir => $opt_origin_cachedir if $opt_origin_cachedir;
    push @dest_opts, -cachedir => $opt_dest_cachedir if $opt_dest_cachedir;
    push @origin_opts, -pathlist => $opt_origin_pathlist if $opt_origin_pathlist;
    push @dest_opts, -pathlist => $opt_dest_pathlist if $opt_dest_pathlist;

    $self->{-originrepo}=$origin_class->new(
	-default_opt=>$ORIGINPATH,
	@ORIGIN_DEFAULT_OPTS,
	$opt_destmode ? () : @defrepo_opts,
	@origin_opts,
	@{$self->{-originopts}},
	);
    $self->{-destrepo}=$dest_class->new(
	-default_opt=>$DESTPATH,
	@DEST_DEFAULT_OPTS,
	$opt_destmode ? @defrepo_opts : (),
	@dest_opts,
	@{$self->{-destopts}},
	);

    $self->{-destmode}=$opt_destmode;
    my $def_repo = $opt_destmode ? $self->{-destrepo} : $self->{-originrepo};
    if ($opt_taskfile) {
	open (my $fn, '<', $opt_taskfile) or die "can't open file $opt_taskfile\n";
	while (my $line=<$fn>) {
	    my @list = map{$def_repo->expand_glob($_)} &Source::Repository::__line2array($line);
	    next if scalar @list == 0;
	    if (scalar @list == 1) {
		push @{$self->{'-packages'}}, @list;
	    } elsif ($list[0] eq '@FOREACH:') {
		shift @list;
		push @{$self->{'-packages'}}, @list;
	    } else {
		# transaction
		push @{$self->{'-packages'}}, [@list];
	    }
	}
	close($fn);
    }

    push @{$self->{'-packages'}}, $opt_transaction ? [@ARGV] : @ARGV;
    push @{$self->{'-packages'}}, $def_repo->names() if $opt_all;
    $self->{-resourceobj}=RPM::Source::Shared::ResourceContainer->new();
    $self->{-comparator}=$comparator_class->new(
	-epoch_comparator => $epoch_comparator,
	-version_comparator=>$version_comparator,
	-release_comparator=>$release_comparator);
    return $self;
}

sub usage {
    print "usage: $0 [ options ] [ list of origin rpms/names ]

Example:
upgrade Sisyphus java packages using jpackage
originmass --dest /var/ftp/pub/Sisyphus/files/SRPMS --origin /var/ftp/pub/jpackage/6.0/generic/free/SRPMS

General options:
	-h, --help
	-v, --verbose (can be specified multiple times)
	-q, --quiet

Mass processing options:
	--origin	/path1/to/foreign|origin/src/RPMS:/other/path2
	--dest	/path1/to/altlinux|dest/src/RPMS:/other/path1:/other/path2
	--skip-on-error	ignore individual package failures
	--strategy 'strategy'	of mass operations (default is $strategy).
";&_print_strategies();

    foreach my $usage(@EXT_OPTS_USAGE) {
	if (ref $usage ne 'CODE') {
	    die "Oops! not a code: $usage ".(ref $usage eq 'SCALAR' ? $$usage : '');
	} else {
	    print "\n";
	    &$usage;
	}
    }
    print STDERR "Mass:usage: EXT_LONGOPT_CLASSES: ",join(', ',@EXT_LONGOPT_CLASSES),"\n" if $verbose and $verbose>6;
    map {$_->longopt_usage} @built_in_containers, @EXT_LONGOPT_CLASSES;
}

sub _print_strategies {
    print "	Available strategies: 
";
    my %main_name=map {$STRATEGYMAP{$_}=>$_} keys (%STRATEGYDESCR);
    my %synonym=map {$_=>[]} keys (%STRATEGYDESCR);
    foreach my $sname (sort keys(%STRATEGYMAP)) {
	my $description=$STRATEGYDESCR{$sname};
	my $mname=$main_name{$STRATEGYMAP{$sname}};
	push@{$synonym{$mname}}, $sname if $mname and not defined $description;
    }
    foreach my $sname (sort keys(%STRATEGYMAP)) {
	my $description=$STRATEGYDESCR{$sname};
	if (defined $description) {
	    print "\t\t$sname\t- $description\n";
	    my @synonyms = @{$synonym{$sname}};
	    print "\t\t\t  Synonyms: ".join(',',@synonyms)."\n" if @synonyms;
	}
    }
}

sub originrepo {
    my ($self)=@_;
    return $self->{-originrepo};
}

sub destrepo {
    my ($self)=@_;
    return $self->{-destrepo};
}

sub comparator {
    my ($self)=@_;
    return $self->{-comparator};
}

sub process {
    my ($self)=@_;
    foreach my $nameobj (@{$self->{'-packages'}}) {
	if (ref $nameobj) {
	    $self->_process_transaction($nameobj);
	} else {
	    $self->_process_name($nameobj);
	}
    }
}

sub _process_name {
    my ($self, $name)=@_;
    my $pkgpair=$self->_match_and_filter($name);
    return unless $pkgpair;
    $self->_process_pair_wrapper($pkgpair);
}

sub _process_transaction {
    my ($self, $transaction_ptr)=@_;
    my $transaction_name=$transaction_ptr->[0];
    my $is_in_transaction=0;
    my $transaction_count=0;
    foreach my $name (@$transaction_ptr) {
	my $pkgpair=$self->_match_and_filter($name);
	next if not $pkgpair;
	if (not $is_in_transaction) {
	    $self->init_transaction($transaction_name);
	    $is_in_transaction=1;
	}
	$self->_process_pair_wrapper($pkgpair, $transaction_name, ++$transaction_count);
    }
    $self->done_transaction($transaction_name,$transaction_count) if $is_in_transaction;
}

sub _match_and_filter {
    my ($self,$input)=@_;
    Carp::confess "Source::Repository::Mass::new_pair: undefined input" unless defined $input;
    my $resourceobj=$self->{-resourceobj};
    my ($destname, $originname, $dest, $origin);
    my $originrepo=$self->{-originrepo};
    my $destrepo=$self->{-destrepo};

    if ($self->{-destmode}) {
	if ($destrepo && $destrepo->is_pkg($input)) {
	    $dest=$destrepo->file2pkg($input);
	    $destname=$dest->name;
	} else {
	    $destname=$input;
	}
    } else {
	if ($originrepo && $originrepo->is_pkg($input)) {
	    $origin=$originrepo->file2pkg($input);
	    $originname=$origin->name;
	} else {
	    $originname=$input;
	}
    }
    if ($self->{-destmode}) {
	$originname=$self->_get_matcher()->dest2origin($destname,$resourceobj);
    } else {
	$destname=$self->_get_matcher()->origin2dest($originname,$resourceobj);
    }

    
    return if Source::Repository::Filter->namefilter($originname,$destname,$resourceobj);
    
    if ($verbose > 2) {
	print STDERR "Source::Repository::Matcher: new_pair($input): origin=$originname dest=$destname\n";
	print STDERR "Source::Repository::Matcher: originrepo=".ref($originrepo)." destrepo=".ref($destrepo)."\n";
	print STDERR "Source::Repository::Matcher: origin=".ref($origin)." dest=".ref($dest)."\n";
    }
    my @pair_args=(
	-originrepo => $originrepo,
	-destrepo => $destrepo,
	-originname => $originname,
	-destname => $destname,
	);
    push @pair_args, -origin => $origin if $origin;
    push @pair_args, -dest => $dest if $dest;
    my $pkgpair=Source::Package::Pair->new(@pair_args);

    # TODO: put it into Mass::PkgFilter. 
    # Mass::PkgFilter should be priority container as is_strategy_allow is rather heavy 
    # - should not be first. priorities should be: PREORIGIN PREDEST ORIGIN DEST
    return unless $self->is_strategy_allow($pkgpair);
    return if Source::Repository::Mass::PkgFilter->pkgfilter($pkgpair,$resourceobj,$self);

    Source::Repository::Filter->postnamefilter($originname,$destname,$resourceobj);
    Source::Repository::Mass::PkgFilter->postpkgfilter($pkgpair,$resourceobj,$self);
    return $pkgpair;
}

sub _get_matcher {
    my $self=shift;
    my $matcher_obj=$self->{-matcher};
    unless ($matcher_obj) {
	$matcher_obj = $matcher_class->new(
	-originrepo => $self->{-originrepo}, 
	-destrepo => $self->{-destrepo}, 
	-destmode=>$opt_destmode,
	);
	$self->{-matcher}=$matcher_obj;
    }
    return $matcher_obj;
}

sub init_transaction {
    my ($self,$transaction_name)=@_;
}

sub done_transaction {
    my ($self,$transaction_name)=@_;
}

sub _process_pair_wrapper {
    my ($self,$pkgpair,$transaction_name,$transaction_count)=@_;
    eval {
	$self->process_pair($pkgpair,$transaction_name,$transaction_count);
    };
    if ($@) {
	die 'FAILED (Mass): '.$pkgpair->originname.": $@

You can ignore this error with --skip-on-error option\n" unless $opt_skip_on_error;
	print STDERR 'FAILED (Mass): '.$pkgpair->originname.": $@";
    };
}

###############################################################
# The method process_pair should be implemented in subclasses;
# disable the prototype implementation completely to allow 
# multiple inheritance tricks with Transform::{Embedded,PassThrough}
###############################################################
#sub process_pair {
#    my ($self,$pkgpair,$transaction_name,$transaction_count)=@_;
#    print STDERR "process_pair: $self\n";
#    ...
#}

sub is_strategy_allow {
    my ($self,$pkgpair)=@_;
    my $strategy_allow=$STRATEGYMAP{$strategy};
    unless ($strategy_allow) {
	&_print_strategies();
	die "strategy $strategy not found.\n" ;
    }
    ## TODO: ABSTRACTIZE AS FILTERS
    my $originname=$pkgpair->originname;
    my $destname=$pkgpair->destname;
    return 1 if &{$strategy_allow}($self, $pkgpair->origin, $pkgpair->dest, $originname, $destname);
    return 0;
}

#------------------------------------------

#------------------------------------------
#    private strategies
#------------------------------------------

sub help_strategy {
    &_print_strategies();
    exit(1);
}

sub origin2dest {
    my ($self,$originname, $destname,$origin, $dest)=@_;
    print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1;
}

sub origin2destS {
    my ($self,$originname, $destname,$origin, $dest)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1;
}

sub destup {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1 if $dest and $self->comparator->cmp($origin,$dest)>0;
    return 0;
}

sub destupS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1 if $dest and $self->comparator->cmp($origin,$dest)>0;
    return 0;
}

sub newer {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $originname not found.\n" unless $dest;
    return 1 if not $dest or $self->comparator->cmp($origin,$dest)>0;
    return 0;
}

sub newerS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $originname not found.\n" unless $dest;
    return 1 if not $dest or $self->comparator->cmp($origin,$dest)>0;
    return 0;
}

sub both {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $originname not found.\n" unless $dest;
    return 1 if $dest;
    return 0;
}


sub bothS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $originname not found.\n" unless $dest;
    return 1 if $dest;
    return 0;
}

sub originskipdest {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    print STDERR "found dest RPM $destname for $originname.\n" if $dest;
    return 0 if $dest;
    return 1;
}

sub originskipdestS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "found dest RPM $destname for $originname.\n" if $dest;
    return 0 if $dest;
    return 1;
}

sub newolder {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 1 unless $origin;
    #print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1 if $dest and $self->comparator->cmp($origin,$dest)<0;
    return 0;
}

sub older {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1 if $dest and $self->comparator->cmp($origin,$dest)<0;
    return 0;
}

sub olderS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1 if $dest and $self->comparator->cmp($origin,$dest)<0;
    return 0;
}

sub destskiporigin {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: dest RPM $destname not found.\n" unless $dest;
    return 0 unless $dest;
    print STDERR "found origin RPM $originname for $destname.\n" if $origin;
    return 0 if $origin;
    return 1;
}

sub destskiporiginS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: dest RPM $destname not found.\n" unless $dest;
    return 0 unless $dest;
    #print STDERR "found origin RPM $originname for $destname.\n" if $origin;
    return 0 if $origin;
    return 1;
}

## ########
# release up; version is the same, something other (release) is different
## ########
sub relup {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1 if $dest and $dest->version eq $origin->version and $self->comparator->cmp($origin,$dest)>0;
    return 0;
}

## ########
# version up; versions are different
## ########
sub verup {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1 if $dest and $dest->version ne $origin->version and $self->comparator->cmp($origin,$dest)>0;
    return 0;
}

#------------------------------------------

1;
