package Source::Repository::Tarball::CPAN;

use strict;
use warnings;

use File::Basename;
use Source::Repository::Tarball;
use Source::Package::Comparators::Raw;

our @ISA=qw/Source::Repository::Tarball/;

our $VERSION=0.002;

our $verbose;

# protected, can be overriden in sublasses
sub _get_name_version_format_from_filename {
   my ($self,$item)=@_;
   return &_cpan_guess_pkg_name_version($item);
}



#------------------------------------------------------
# shared static CPAN maintainance code
#------------------------------------------------------

#------------------------------------------------------
my %KNOWN_NUMERICS=map{$_=>1} qw!
ACME-Error-31337
SWISH-3
Acme-123
Acme-1337
Acme-24
Acme-6502
Acme-Test-42
ACME-YAPC-NA-2012
Chess-960
Class-DBI-Frozen-301
Games-2048
Lingua-31337
Perl-Dist-Strawberry-BuildPerl-5123
Perl-Dist-WiX-BuildPerl-5123
Pod-Simple-31337
POSIX-1003
POSIX-2008
tinyperl-1.0-580
!;
my %KNOWN_ALPHABETIC=map{$_=>1} qw!
BGPmon-Analytics-db
BGPmon-Archiver
BGPmon-Client
BGPmon-core
BGPmon-CPM
BGPmon-CPM-Prefix-Finder
BGPmon-Filter
BGPmon-Validate
Math-GoldenBigMath
Math-MatrixReal-Ext1
PDF-CreateSimple
!;
my %KNOWN_DOT_SEPARATOR=map{$_=>1} qw!
AI-ExpertSystem-Simple
FLAT-Legacy-FA
!;
my %KNOWN_EXCEPTIONS=(
    'Fraction-v.53b' => ['Fraction', '0.53b'],
    'WWW-TMDB-API0.03' => ['WWW-TMDB-API', '0.03'],
    'WWW-TMDB-API0.04' => ['WWW-TMDB-API', '0.04'],
    'XMS-MotifSetv1.0' => ['XMS-MotifSet','v1.0'],
    'Win32-TaskScheduler2.0.3' => ['Win32-TaskScheduler','2.0.3'],
    'Win32-Exchange_v0.046a' => ['Win32-Exchange','v0.046a']
    );

#------------------------------------------------------

sub _cpan_guess_pkg_name_version {
    my ($pkg)=@_;
    unless ($pkg=~s/\.(?:tar.gz|tar.bz2|tgz|zip)$//) {
	warn "WARN: cpan_guess_pkg_name_version: not a package: $pkg\n";
	return;
    }
    my $module=basename($pkg);
    my $ver;
    my $separator;
    if ($module=~s/-([^-]+)$//) {
	$ver=$1;
	$separator='-';
	print STDERR "cpan_guess:1: sep=$separator mod=$module ver=$ver\n" if $verbose;
	#HTML-WikiConverter-DokuWikiFCK-0.32_Feb-27-2013
	if ($ver=~/^20[12]\d$/ and $module=~s/-([\d\.]+_\w+[-\.]\d\d)$//) {
	    $ver=$1.$separator.$ver;
	} elsif (not $KNOWN_NUMERICS{$module} and $module=~s/-([\d\.]+)$//) {
	    my $prefix=$1;
	    if ($prefix=~/\d\.\d/ or $KNOWN_ALPHABETIC{$module}) {
		$ver="$prefix$separator$ver";
	    } else {
		$module.=$separator.$prefix;
		warn "cpan_guess_pkg_name_version: WARN: name_with_digits:$pkg $module $ver\n";
	    }
	}
    } elsif ($module=~s/[a-zA-Z]\.(\d[\d\.]+)$//) {
	$ver=$1;
	$separator='.';
	print STDERR "cpan_guess:2: sep=$separator mod=$module ver=$ver\n" if $verbose;
    } elsif ($module=~s/_([^_]+)(?:_(?:stable|alpha|bis|withoutworldwriteables|source|OpenSource))?$//) {
	$ver=$1;
	$separator='_';
	print STDERR "cpan_guess:3: sep=$separator mod=$module ver=$ver\n" if $verbose;
    } elsif ($module=~s/\.(\d[\d\.]+)$//) {
	$ver=$1;
	$separator='.';
	print STDERR "cpan_guess:4: sep=$separator mod=$module ver=$ver\n" if $verbose;
    } else {
	return;
    }

    if ($KNOWN_EXCEPTIONS{$pkg}) {
	($module, $ver)=@{$KNOWN_EXCEPTIONS{$pkg}};
    } elsif ($ver=~/^(\D\D+)\.([\d\.]+)$/ and $KNOWN_DOT_SEPARATOR{$module.'-'.$1}) {
	$module.=$separator.$1;
	$ver=$2;
    } elsif ($ver=~/^(\D\D+)_V([\d\.]+)$/) {
	$module.=$separator.$1;
	$ver=$2;
    }

    # MARC-Lint_1.47 - todo: merge with above code
    if ($ver=~s/([A-Za-z]+)_(?=\d)//) {
	$module.='-'.$1;
    }
    $ver=&_cpan_trim_version($ver,$module);
    return ($module,$ver);
}

my %DASH_TO_SEP =qw/
SOAP-MIME .
/;

sub _cpan_trim_version {
    my ($ver,$module)=@_;
    # (PERL_SUFFIX_TRIM) code
    #$ver=~s/^[vV]//;
    $ver=~s/^V//;
    $ver=~s/[_-](?:stable|alpha|bis|withoutworldwriteables|source|OpenSource)$//i;
    $ver=~s/-(fix|FIXED|beta6_02|beta13|win32)$/_$1/;
    $ver=~s/(?<=\d)beta[123]$//;
    if ($ver=~/^(.*)-(\d+.*)$/) {
	my $pref=$1;
	my $suf=$2;
	my $dash_to_sep=$DASH_TO_SEP{$module};
	# TODO: list explicitly all cpanbuilder rpms that use _ and change defaults to .
	$dash_to_sep ||= '_';
	$ver=$pref.$dash_to_sep.$suf;
    }
    $ver=~s/^[vV]//;
    return $ver;
}


my %KNOWN_BADVER_DROP=qw!
article	fetcher
Bundle	FinalTest
Bundle	FinalTest2
CGI-Application-Muto	withoutworldwriteables
Device-Velleman-K8055	Client
Dancer-OO	depfix
DBIx-HTMLView	LATEST
Email-IsFree	CN
Geo-GoogleEarth-Document	modules2
IPC	SysV
Lingua-TreeTagger	Filter
link		NCBI
manish		db
Math-Expr	LATEST
NDFD-Weather	Report
NET		NETCONF
perl		mlb
Prolog		alpha
RSS-Video	Google
Sendmail	M4.0.29a
Spreadsheet-ParseExcel	Assist
Spreadsheet-WriteExcel	WebPivot2
tinyperl-1.0-580	win32
Tk		GraphMan
WebPresence	Profile
XML		DB
!;
my %KNOWN_NOVERSION_DROP=map{$_=>1} qw!
bid
CECALA
CGISession
DataCube
Diamond
EasyDB.0.5b2
examples
FileKGlob
frogbak
FServer
JomonTest
log
lot
Notes
perlSGML.1997Sep18
PGForth1.0
PGForth1.3
rms
rpn
Saraltest
smg
swd
swig1.1p5
Syringe
tic
TimeConvert0.5
UDPServersAndClients
!;

sub __cpan_load_process_srcpath {
    my ($SRC, $VER, $srcpath, $is_verbose)=@_;
    my $pkg=basename($srcpath);
    return if $pkg=~/\.pm\.(?:gz|bz2)$/;
    my ($module,$ver)=&Source::Repository::Tarball::CPAN::_cpan_guess_pkg_name_version($pkg);
    if (not defined $module) {
	$pkg=~s/\.(?:tar.gz|tar.bz2|tgz|zip)$//;
	print STDERR "STRANGE2: $pkg\n" if $is_verbose and not $KNOWN_NOVERSION_DROP{$pkg};
	return;
    }

    return if $KNOWN_BADVER_DROP{$module} and $KNOWN_BADVER_DROP{$module} eq $ver;
    eval {
	if (not $SRC->{$module} or &Source::Package::Comparators::Raw::perl_version_cmp($VER->{$module},$ver)<=0) {
		$SRC->{$module}=$srcpath;
		$VER->{$module}=$ver;
	}
    };
    if ($is_verbose and $@) {
	my $msg=$@;
	$msg=~s, at /usr.*$,,;
	print STDERR "pkg=$pkg module=$module oldver=$VER->{$module} newver=$ver: $msg\n";
    }
}

sub _cpan_load_02packages_details_from_FH {
    my ($fh, $is_verbose) =@_;
    my %SRC;
    my %VER;
    while (<$fh>) {
	last if /^\s*$/;
    }
    while (<$fh>) {
	my @pklist=split(/\s+/);
	my $srcpath=$pklist[2];
	&__cpan_load_process_srcpath(\%SRC, \%VER, $srcpath, $is_verbose);
    }
    return (\%SRC,\%VER);
}

sub _cpan_load_01modules_index_html_from_FH {
    my ($fh, $is_verbose) =@_;
    my %SRC;
    my %VER;
    local $_;
    while (<$fh>) {
	chomp;
	&__cpan_load_process_srcpath(\%SRC, \%VER, $_, $is_verbose);
    }
    return (\%SRC,\%VER);
}

1;
